﻿using AIWolf.Lib;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection.Metadata.Ecma335;
using System.Threading;
using WhiteWolf.ActionPlan;
using WhiteWolf.Arrangement;
using WhiteWolf.Guess;
using WhiteWolf.Player;
using WhiteWolf.Talk;
using static WhiteWolf.ContestAgent.Statistics;

namespace WhiteWolf.ContestAgent
{
    public class ContestUdonPlayer : AbstractWhiteWolfPlayer
    {

        public override string Name => "Udon";

        private IGuesser guesser;
        protected override IGuesser Guesser => guesser;

        private IActionPlanner actionPlanner;
        protected override IActionPlanner ActionPlanner => actionPlanner;

        ITalkGenerator talkGenerator;
        protected override ITalkGenerator TalkGenerator => talkGenerator;

        ITalkGenerator whisperGenerator = new DummyTalkGenerator();
        protected override ITalkGenerator WhisperGenerator => whisperGenerator;

        private IGameInfo latestGameInfo;

        private LearnData LearnData { get; set; }


        public ContestUdonPlayer()
        {
        }

        public override void Initialize(IGameInfo gameInfo, IGameSetting gameSetting)
        {
            try
            {
                // 初回ゲーム開始時に学習オブジェクト初期化
                if (LearnData == null)
                {
                    LearnData = new LearnData(gameInfo.AgentList);
                    guesser = new UdonGuesser(LearnData);
                    actionPlanner = new UdonActionPlanner((UdonGuesser)guesser, LearnData);
                    talkGenerator = new UdonTalkGenerator(LearnData);
                }

                TotallyGameInfo = new TotalityGameInfo(gameInfo, gameSetting);
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Initialize Error.");
            }
        }

        public override void Update(IGameInfo gameInfo)
        {
            try
            {
                latestGameInfo = gameInfo;
                TotallyGameInfo.Update(gameInfo);

                // 村人でも役職を騙るエージェントがCO経由せずに判定出ししていたら、COと判定を取り消す
                var yabaiJudge = TotallyGameInfo.SeerJudges
                    .Concat(TotallyGameInfo.MediumJudges)
                    .Where(judge => TotallyGameInfo.ComingOuts.Any(co => judge.JudgeTalk.Equals(co)) && LearnData.Statistics[judge.Agent][Role.VILLAGER].FakeCO > 0);
                yabaiJudge.ToList().ForEach(judge => judge.Cancel(judge.JudgeTalk));
                TotallyGameInfo.ComingOuts.Where(co => yabaiJudge.Any(judge => co.ComingOutTalk.Equals(judge))).ToList().ForEach(co => co.Cancel(co.ComingOutTalk));
                if (yabaiJudge.Any())
                {
                    TotallyGameInfo.RebuildViewPoint();
                }

                LatestTotalGuess = Guesser.GetGuess(TotallyGameInfo);
                LatestActionPlan = ActionPlanner.GetPlan(TotallyGameInfo, LatestTotalGuess);
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Update Error.");
            }
        }

        public override void Finish()
        {
            try
            {
                var agentList = latestGameInfo.AgentList;
                var winner = latestGameInfo.RoleMap.Any(map => latestGameInfo.AliveAgentList.Contains(map.Key) && map.Value == Role.WEREWOLF) ? Team.WEREWOLF : Team.VILLAGER;
                var roleMap = latestGameInfo.RoleMap;

                roleMap.ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].GameCount++);
                roleMap.Where(map => map.Value.GetTeam() == winner).ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].WinCount++);
                roleMap.Where(map => map.Value.GetTeam() != winner).ToList().ForEach(map => LearnData.Statistics[map.Key][map.Value].LoseCount++);

                TotallyGameInfo.ComingOuts.Where(co => co.ComingOutTalk.Day == 1).ToList().ForEach(co =>
                {
                    if (co.Role == Role.SEER) LearnData.Statistics[co.Agent][roleMap[co.Agent]].SeerCO++;
                    if (co.Role == Role.MEDIUM) LearnData.Statistics[co.Agent][roleMap[co.Agent]].MediumCO++;
                    if (roleMap[co.Agent] != co.Role && co.Role != Role.BODYGUARD) LearnData.Statistics[co.Agent][roleMap[co.Agent]].FakeCO++;
                });

                TotallyGameInfo.SeerJudges.Where(judge => judge.JudgeTalk.Day == 1 && judge.Result == Species.HUMAN).ToList()
                    .ForEach(judge => LearnData.Statistics[judge.Agent][roleMap[judge.Agent]].FirstDivineIsWhite++);
                TotallyGameInfo.SeerJudges.Where(judge => judge.JudgeTalk.Day == 1 && judge.Result == Species.WEREWOLF).ToList()
                    .ForEach(judge => LearnData.Statistics[judge.Agent][roleMap[judge.Agent]].FirstDivineIsBlack++);

                // 発話はNOPかANDの先頭のみ見る
                TotallyGameInfo.DayGameInfo.SelectMany(day => day.Value.TalkList).ToList().ForEach(talk =>
                {
                    if (talk.Content.Operator == Operator.NOP) UpdateTalkStatistics(talk.Agent, roleMap[talk.Agent], talk, talk.Content);
                    if (talk.Content.Operator == Operator.AND && talk.Content.ContentList[0].Operator == Operator.NOP) UpdateTalkStatistics(talk.Agent, roleMap[talk.Agent], talk, talk.Content.ContentList[0]);
                });

                foreach (var day in TotallyGameInfo.DayGameInfo.Where(day => day.Value.VoteList.Any()))
                {
                    var voteMap = day.Value.VoteList[0].VoteMap;
                    var ReceiveMap = voteMap.GroupBy(map => map.Value).ToList();
                    foreach (var vote in voteMap)
                    {
                        LearnData.Statistics[vote.Key][roleMap[vote.Key]].Vote++;

                        int cnt = ReceiveMap.SingleOrDefault(g => g.Key.Equals(vote.Value))?.Count() ?? 0;
                        int max = ReceiveMap.Where(g => !g.Key.Equals(vote.Key)).Max(g => g.Count());
                        if (cnt == 1 && voteMap.Any(map => !map.Value.Equals(vote.Key)))
                        {
                            LearnData.Statistics[vote.Key][roleMap[vote.Key]].VoteToNoReceive++;
                        }
                        if (cnt == max)
                        {
                            LearnData.Statistics[vote.Key][roleMap[vote.Key]].VoteToMax++;
                        }
                    }
                }

                UpdatePlayerMeta5();
                UpdatePlayerMeta15();
            }
            catch (Exception ex)
            {
                //Console.WriteLine("Finish Error.");
            }
        }


        private void UpdateTalkStatistics(Agent agent, Role role, IUtterance talk, Content content)
        {
            switch (content.Topic)
            {
                case Topic.Over:
                case Topic.Skip:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.OverAndSkip]++;
                    break;
                case Topic.DIVINED:
                case Topic.IDENTIFIED:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Judge]++;
                    break;
                case Topic.COMINGOUT:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.ComingOut]++;
                    break;
                case Topic.VOTE:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Vote]++;
                    break;
                case Topic.ESTIMATE:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Estimate]++;
                    break;
                default:
                    LearnData.Statistics[agent][role].Talk[talk.Day][talk.Turn][(int)TalkType.Other]++;
                    break;
            }
        }

        class DummyTalkGenerator : ITalkGenerator
        {
            public string GetTalkText(ITotalityGameInfo gameInfo, ITotalGuess guess, IActionPlan actionPlan)
            {
                return Content.OVER.Text;
            }
        }

        private void UpdatePlayerMeta5()
        {
            if (TotallyGameInfo.GameSetting.PlayerNum != 5) return;
            var roleMap = latestGameInfo.RoleMap;
            
            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Turn == 0 && talk.Text.Contains("REQUEST"))
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Camellia]++;
                    Console.WriteLine($"{talk.Agent} is Camellia.");
                });

            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Turn == 0 && talk.Content.Topic == Topic.COMINGOUT && talk.Content.Role == Role.VILLAGER)
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Tomato]++;
                    Console.WriteLine($"{talk.Agent} is Tomato.");
                });

            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Operator == Operator.NOP && talk.Content.Topic == Topic.ESTIMATE && roleMap[talk.Agent] != Role.WEREWOLF)
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.wasabi]++;
                    Console.WriteLine($"{talk.Agent} is wasabi.");
                });
            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Content.Operator == Operator.NOP && talk.Content.Topic == Topic.ESTIMATE && talk.Content.Role == Role.WEREWOLF && roleMap[talk.Agent] == Role.WEREWOLF)
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.otsuki]++;
                    Console.WriteLine($"{talk.Agent} is otsuki or wasabi.");
                });

            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Text.Contains("DAY") && talk.Text.Contains("BECAUSE"))
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Sashimi]++;
                    Console.WriteLine($"{talk.Agent} is Sashimi.");
                });

            TotallyGameInfo.DayGameInfo[1].TalkList
                .Where(talk => talk.Text.Contains("AND (VOTE"))
                .ToList()
                .ForEach(talk => {
                    LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Sashimi]++;
                    Console.WriteLine($"{talk.Agent} is Sashimi.");
                });

            foreach (var agent in TotallyGameInfo.AgentList)
            {
                if (roleMap[agent] != Role.VILLAGER) continue;
                if (TotallyGameInfo.DayGameInfo[1].TalkList.Count(talk => talk.Agent.Equals(agent) && talk.Content.Topic == Topic.VOTE) < 5) continue;

                // takeda, Daisyo, HALU, Sashimi, Tomo, Cube, Simipu
                LearnData.Statistics[agent][roleMap[agent]].Player[(int)PlayerType.takeda]++;
                Console.WriteLine($"{agent} is takeda or childlen.");
            }

            var seer = roleMap.Single(kv => kv.Value == Role.SEER).Key;

        }

        private void UpdatePlayerMeta15()
        {
            if (TotallyGameInfo.GameSetting.PlayerNum != 15) return;
            var roleMap = latestGameInfo.RoleMap;

            // takeda派生？（takedaそのものは発話パターン違いそう？）
            foreach (var agent in TotallyGameInfo.AgentList)
            {
                if (roleMap[agent] != Role.VILLAGER) continue;
                if (TotallyGameInfo.DayGameInfo[1].TalkList.Count(talk => talk.Agent.Equals(agent) && talk.Content.Topic == Topic.VOTE) < 5) continue;

                // takeda, HALU, Sashimi, Tomo, Cube, Simipu
                LearnData.Statistics[agent][roleMap[agent]].Player[(int)PlayerType.takeda]++;
                Console.WriteLine($"{agent} is takeda or childlen.");
            }

            // Tomato 狼で村CO
            foreach (var agent in TotallyGameInfo.AgentList)
            {
                if (roleMap[agent] != Role.WEREWOLF) continue;
                if (!TotallyGameInfo.DayGameInfo[1].TalkList.Any(talk => talk.Agent.Equals(agent) && talk.Content.Topic == Topic.COMINGOUT && talk.Content.Role == Role.VILLAGER)) continue;
                LearnData.Statistics[agent][roleMap[agent]].Player[(int)PlayerType.Tomato]++;
                Console.WriteLine($"{agent} is Tomato.");
            }

            // Camellia
            foreach (var co in TotallyGameInfo.ComingOuts.Where(co => co.Role == Role.BODYGUARD))
            {
                LearnData.Statistics[co.Agent][roleMap[co.Agent]].Player[(int)PlayerType.Camellia]++;
                Console.WriteLine($"{co.Agent} is Camellia.");
            }
            foreach (var talk in TotallyGameInfo.DayGameInfo[1].TalkList.Where(talk => talk.Text.Contains("REQUEST") && talk.Text.Contains("DIVINATION")))
            {
                if (roleMap[talk.Agent] != Role.VILLAGER) continue;
                LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Camellia]++;
                Console.WriteLine($"{talk.Agent} is Camellia.");
            }

            // Sashimi
            foreach (var talk in TotallyGameInfo.DayGameInfo[1].TalkList.Where(talk => talk.Text.Contains("AND (VOTE")))
            {
                LearnData.Statistics[talk.Agent][roleMap[talk.Agent]].Player[(int)PlayerType.Sashimi]++;
                Console.WriteLine($"{talk.Agent} is Sashimi.");
            }

            var seer = roleMap.Single(kv => kv.Value == Role.SEER).Key;

        }


    }
}
